/*
 * Copyright 2016 Intel Corporation
 * Copyright 2016 Broadcom
 * Copyright 2020 Collabora, Ltd.
 * Copyright 2024 Alyssa Rosenzweig
 * Copyright 2025 Lima Project
 * SPDX-License-Identifier: MIT
 */

#pragma once

#include <inttypes.h>
#include <stdio.h>
#include "util/bitpack_helpers.h"
#include "util/half_float.h"
#define FILE_TYPE FILE
#define CONSTANT_ const
#define GLOBAL_

#define __gen_unpack_float(x, y, z) uif(__gen_unpack_uint(x, y, z))
#define __gen_unpack_half(x, y, z)                                             \
   _mesa_half_to_float(__gen_unpack_uint(x, y, z))

static inline uint64_t
__gen_unpack_uint(CONSTANT_ uint32_t *restrict cl, uint32_t start, uint32_t end)
{
   uint64_t val = 0;
   const int width = end - start + 1;
   const uint64_t mask =
      (width == 64) ? ~((uint64_t)0) : ((uint64_t)1 << width) - 1;

   for (unsigned word = start / 32; word < (end / 32) + 1; word++) {
      val |= ((uint64_t)cl[word]) << ((word - start / 32) * 32);
   }

   return (val >> (start % 32)) & mask;
}

static inline uint64_t
__gen_pack_unorm16(float f, uint32_t start, uint32_t end)
{
   return util_bitpack_uint(float_to_ushort(f), start, end);
}

static inline float
__gen_unpack_unorm16(CONSTANT_ uint32_t *restrict cl, uint32_t start, uint32_t end)
{
   return ushort_to_float(__gen_unpack_uint(cl, start, end));
}

static inline uint64_t
__gen_unpack_sint(CONSTANT_ uint32_t *restrict cl, uint32_t start, uint32_t end)
{
   int size = end - start + 1;
   int64_t val = __gen_unpack_uint(cl, start, end);

   return util_sign_extend(val, size);
}

static inline float
__gen_unpack_ulod(const uint32_t *restrict cl, uint32_t start, uint32_t end)
{
   uint32_t u = __gen_unpack_uint(cl, start, end);

   return ((float)u) / 16.0;
}

static inline float
__gen_unpack_slod(const uint32_t *restrict cl, uint32_t start, uint32_t end)
{
   int32_t u = __gen_unpack_sint(cl, start, end);

   return ((float)u) / 16.0;
}

static inline uint64_t
__gen_to_groups(uint32_t value, uint32_t group_size, uint32_t length)
{
   /* Zero is not representable, clamp to minimum */
   if (value == 0)
      return 1;

   /* Round up to the nearest number of groups */
   uint32_t groups = DIV_ROUND_UP(value, group_size);

   /* The 0 encoding means "all" */
   if (groups == ((uint64_t)1) << length)
      return 0;

   /* Otherwise it's encoded as the identity */
   assert(groups < (1u << length) && "out of bounds");
   assert(groups >= 1 && "exhaustive");
   return groups;
}

static inline uint64_t
__gen_from_groups(uint32_t value, uint32_t group_size, uint32_t length)
{
   return group_size * (value ? value : (1 << length));
}

#define lima_pack(dst, T, name)                                                \
   for (struct LIMA_##T name = {LIMA_##T##_header},                            \
                       *_loop_count = (GLOBAL_ void *)((uintptr_t)0);          \
        (uintptr_t)_loop_count < 1; (                                          \
           {                                                                   \
              LIMA_##T##_pack((GLOBAL_ uint32_t *)(dst), &name);               \
              _loop_count = (GLOBAL_ void *)(((uintptr_t)_loop_count) + 1);    \
           }))

#define lima_unpack(fp, src, T, name)                                          \
   struct LIMA_##T name;                                                       \
   LIMA_##T##_unpack(fp, (CONSTANT_ uint8_t *)(src), &name)

#define lima_print(fp, T, var, indent) LIMA_##T##_print(fp, &(var), indent)

static inline void
lima_merge_helper(uint32_t *dst, const uint32_t *src, size_t bytes)
{
   assert((bytes & 3) == 0);

   for (unsigned i = 0; i < (bytes / 4); ++i)
      dst[i] |= src[i];
}

#define lima_merge(packed1, packed2, type)                                      \
   lima_merge_helper((packed1).opaque, (packed2).opaque, LIMA_##type##_LENGTH)

#if defined(NDEBUG)
#define lima_genxml_validate_bounds(a, b, c)
#define lima_genxml_validate_mask(a, b, c, d, e) true
#define lima_genxml_validate_exact(a, b, c, d)   true
#else
static inline void
lima_genxml_validate_bounds(const char *name, uint64_t value, uint64_t bound)
{
   if (unlikely(value >= bound)) {
      fprintf(stderr, "%s out-of-bounds, got 0x%" PRIx64 ", max %" PRIx64 "\n",
              name, value, bound);

      unreachable("Out-of-bounds pack");
   }
}

static inline bool
lima_genxml_validate_mask(FILE *fp, const char *name, const void *cl_,
                         uint32_t index, uint32_t bad_mask)
{
   const uint32_t *cl = (const uint32_t *)cl_;
   uint32_t bad = cl[index] & bad_mask;

   if (bad && fp != NULL) {
      fprintf(
         fp,
         "XXX: Unknown field of %s unpacked at word %u got %X, bad mask %X\n",
         name, index, cl[index], bad);
   }

   return bad == 0;
}

static bool
lima_genxml_validate_exact(FILE *fp, const char *name, uint64_t value,
                          uint64_t exact)
{
   if (value != exact && fp != NULL) {
      fprintf(fp, "XXX: Expected %s to equal %" PRIx64 " but got %" PRIx64 "\n",
              name, value, exact);
   }

   return value == exact;
}

#endif

/* Everything after this is autogenerated from XML. Do not hand edit. */

enum lima_texel_format {
   LIMA_TEXEL_FORMAT_L8 = 9,
   LIMA_TEXEL_FORMAT_A8 = 10,
   LIMA_TEXEL_FORMAT_I8 = 11,
   LIMA_TEXEL_FORMAT_BGR_565 = 14,
   LIMA_TEXEL_FORMAT_BGRA_5551 = 15,
   LIMA_TEXEL_FORMAT_BGRA_4444 = 16,
   LIMA_TEXEL_FORMAT_L8A8 = 17,
   LIMA_TEXEL_FORMAT_L16 = 18,
   LIMA_TEXEL_FORMAT_A16 = 19,
   LIMA_TEXEL_FORMAT_I16 = 20,
   LIMA_TEXEL_FORMAT_RGB_888 = 21,
   LIMA_TEXEL_FORMAT_RGBA_8888 = 22,
   LIMA_TEXEL_FORMAT_RGBX_8888 = 23,
   LIMA_TEXEL_FORMAT_ETC1_RGB8 = 32,
   LIMA_TEXEL_FORMAT_L16_FLOAT = 34,
   LIMA_TEXEL_FORMAT_A16_FLOAT = 35,
   LIMA_TEXEL_FORMAT_I16_FLOAT = 36,
   LIMA_TEXEL_FORMAT_L16A16_FLOAT = 37,
   LIMA_TEXEL_FORMAT_R16G16B16A16_FLOAT = 38,
   LIMA_TEXEL_FORMAT_Z24X8 = 44,
   LIMA_TEXEL_FORMAT_R16G16B16_FLOAT = 47,
   LIMA_TEXEL_FORMAT_Z24S8_RLD = 50,
};

static inline const char *
lima_texel_format_as_str(enum lima_texel_format imm)
{
    switch (imm) {
    case LIMA_TEXEL_FORMAT_L8: return "L8";
    case LIMA_TEXEL_FORMAT_A8: return "A8";
    case LIMA_TEXEL_FORMAT_I8: return "I8";
    case LIMA_TEXEL_FORMAT_BGR_565: return "BGR_565";
    case LIMA_TEXEL_FORMAT_BGRA_5551: return "BGRA_5551";
    case LIMA_TEXEL_FORMAT_BGRA_4444: return "BGRA_4444";
    case LIMA_TEXEL_FORMAT_L8A8: return "L8A8";
    case LIMA_TEXEL_FORMAT_L16: return "L16";
    case LIMA_TEXEL_FORMAT_A16: return "A16";
    case LIMA_TEXEL_FORMAT_I16: return "I16";
    case LIMA_TEXEL_FORMAT_RGB_888: return "RGB_888";
    case LIMA_TEXEL_FORMAT_RGBA_8888: return "RGBA_8888";
    case LIMA_TEXEL_FORMAT_RGBX_8888: return "RGBX_8888";
    case LIMA_TEXEL_FORMAT_ETC1_RGB8: return "ETC1_RGB8";
    case LIMA_TEXEL_FORMAT_L16_FLOAT: return "L16_FLOAT";
    case LIMA_TEXEL_FORMAT_A16_FLOAT: return "A16_FLOAT";
    case LIMA_TEXEL_FORMAT_I16_FLOAT: return "I16_FLOAT";
    case LIMA_TEXEL_FORMAT_L16A16_FLOAT: return "L16A16_FLOAT";
    case LIMA_TEXEL_FORMAT_R16G16B16A16_FLOAT: return "R16G16B16A16_FLOAT";
    case LIMA_TEXEL_FORMAT_Z24X8: return "Z24X8";
    case LIMA_TEXEL_FORMAT_R16G16B16_FLOAT: return "R16G16B16_FLOAT";
    case LIMA_TEXEL_FORMAT_Z24S8_RLD: return "Z24S8_RLD";
    default: return NULL;
    }
}
enum lima_texture_range {
   LIMA_TEXTURE_RANGE_UNORM = 0,
   LIMA_TEXTURE_RANGE_SNORM_HALF = 1,
   LIMA_TEXTURE_RANGE_SNORM = 2,
   LIMA_TEXTURE_RANGE_UNORM_INVERTED = 3,
};

static inline const char *
lima_texture_range_as_str(enum lima_texture_range imm)
{
    switch (imm) {
    case LIMA_TEXTURE_RANGE_UNORM: return "UNORM";
    case LIMA_TEXTURE_RANGE_SNORM_HALF: return "SNORM HALF";
    case LIMA_TEXTURE_RANGE_SNORM: return "SNORM";
    case LIMA_TEXTURE_RANGE_UNORM_INVERTED: return "UNORM INVERTED";
    default: return NULL;
    }
}
enum lima_sampler_dimension {
   LIMA_SAMPLER_DIMENSION_1D = 0,
   LIMA_SAMPLER_DIMENSION_2D = 1,
   LIMA_SAMPLER_DIMENSION_3D = 2,
};

static inline const char *
lima_sampler_dimension_as_str(enum lima_sampler_dimension imm)
{
    switch (imm) {
    case LIMA_SAMPLER_DIMENSION_1D: return "1D";
    case LIMA_SAMPLER_DIMENSION_2D: return "2D";
    case LIMA_SAMPLER_DIMENSION_3D: return "3D";
    default: return NULL;
    }
}
enum lima_mipfilter {
   LIMA_MIPFILTER_NEAREST = 0,
   LIMA_MIPFILTER_LINEAR = 3,
};

static inline const char *
lima_mipfilter_as_str(enum lima_mipfilter imm)
{
    switch (imm) {
    case LIMA_MIPFILTER_NEAREST: return "Nearest";
    case LIMA_MIPFILTER_LINEAR: return "Linear";
    default: return NULL;
    }
}
enum lima_tex_wrap {
   LIMA_TEX_WRAP_REPEAT = 0,
   LIMA_TEX_WRAP_CLAMP_TO_EDGE = 1,
   LIMA_TEX_WRAP_CLAMP = 2,
   LIMA_TEX_WRAP_CLAMP_TO_BORDER = 3,
   LIMA_TEX_WRAP_MIRROR_REPEAT = 4,
   LIMA_TEX_WRAP_MIRROR_CLAMP_TO_EDGE = 5,
   LIMA_TEX_WRAP_MIRROR_CLAMP = 6,
   LIMA_TEX_WRAP_MIRROR_CLAMP_TO_BORDER = 7,
};

static inline const char *
lima_tex_wrap_as_str(enum lima_tex_wrap imm)
{
    switch (imm) {
    case LIMA_TEX_WRAP_REPEAT: return "Repeat";
    case LIMA_TEX_WRAP_CLAMP_TO_EDGE: return "Clamp To Edge";
    case LIMA_TEX_WRAP_CLAMP: return "Clamp";
    case LIMA_TEX_WRAP_CLAMP_TO_BORDER: return "Clamp To Border";
    case LIMA_TEX_WRAP_MIRROR_REPEAT: return "Mirror Repeat";
    case LIMA_TEX_WRAP_MIRROR_CLAMP_TO_EDGE: return "Mirror Clamp To Edge";
    case LIMA_TEX_WRAP_MIRROR_CLAMP: return "Mirror Clamp";
    case LIMA_TEX_WRAP_MIRROR_CLAMP_TO_BORDER: return "Mirror Clamp To Border";
    default: return NULL;
    }
}
enum lima_texture_layout {
   LIMA_TEXTURE_LAYOUT_LINEAR = 0,
   LIMA_TEXTURE_LAYOUT_TILED = 3,
};

static inline const char *
lima_texture_layout_as_str(enum lima_texture_layout imm)
{
    switch (imm) {
    case LIMA_TEXTURE_LAYOUT_LINEAR: return "Linear";
    case LIMA_TEXTURE_LAYOUT_TILED: return "Tiled";
    default: return NULL;
    }
}
struct LIMA_TEXTURE_DESCRIPTOR {
   enum lima_texel_format               texel_format;
   bool                                 reverse_component_order;
   bool                                 swap_rb;
   enum lima_texture_range              range;
   bool                                 has_signed_components;
   uint32_t                             unknown_0;
   uint32_t                             stride;
   uint32_t                             unknown_1;
   bool                                 unnorm_coords;
   bool                                 unknown_2;
   bool                                 cube_map;
   enum lima_sampler_dimension          sampler_dim;
   float                                min_lod;
   float                                max_lod;
   float                                lod_bias;
   uint32_t                             unknown_3;
   bool                                 has_stride;
   enum lima_mipfilter                  mipfilter;
   bool                                 min_img_filter_nearest;
   bool                                 mag_img_filter_nearest;
   enum lima_tex_wrap                   wrap_s;
   enum lima_tex_wrap                   wrap_t;
   enum lima_tex_wrap                   wrap_r;
   uint32_t                             width;
   uint32_t                             height;
   uint32_t                             depth;
   float                                border_red;
   float                                border_green;
   float                                border_blue;
   float                                border_alpha;
   uint32_t                             unknown_4;
   enum lima_texture_layout             layout;
   uint32_t                             unknown_5;
   uint32_t                             mip_0;
   uint32_t                             mip_1;
   uint32_t                             mip_2;
   uint32_t                             mip_3;
   uint32_t                             mip_4;
   uint32_t                             mip_5;
   uint32_t                             mip_6;
   uint32_t                             mip_7;
   uint32_t                             mip_8;
   uint32_t                             mip_9;
   uint32_t                             mip_10;
   uint32_t                             unknown_6;
};

#define LIMA_TEXTURE_DESCRIPTOR_header 0

static inline void
LIMA_TEXTURE_DESCRIPTOR_pack(GLOBAL_ uint32_t * restrict cl,
                             const struct LIMA_TEXTURE_DESCRIPTOR * restrict values)
{
   assert((values->mip_0 & 0x3f) == 0);
   assert((values->mip_1 & 0x3f) == 0);
   assert((values->mip_2 & 0x3f) == 0);
   assert((values->mip_3 & 0x3f) == 0);
   assert((values->mip_4 & 0x3f) == 0);
   assert((values->mip_5 & 0x3f) == 0);
   assert((values->mip_6 & 0x3f) == 0);
   assert((values->mip_7 & 0x3f) == 0);
   assert((values->mip_8 & 0x3f) == 0);
   assert((values->mip_9 & 0x3f) == 0);
   assert((values->mip_10 & 0x3f) == 0);
   lima_genxml_validate_bounds("Texture Descriptor::texel_format", values->texel_format, 0x40ull);
   lima_genxml_validate_bounds("Texture Descriptor::reverse_component_order", values->reverse_component_order, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::swap_rb", values->swap_rb, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::range", values->range, 0x4ull);
   lima_genxml_validate_bounds("Texture Descriptor::has_signed_components", values->has_signed_components, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_0", values->unknown_0, 0x20ull);
   lima_genxml_validate_bounds("Texture Descriptor::stride", values->stride, 0x8000ull);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_1", values->unknown_1, 0x100ull);
   cl[ 0] = util_bitpack_uint(values->texel_format, 0, 5) |
            util_bitpack_uint(values->reverse_component_order, 6, 6) |
            util_bitpack_uint(values->swap_rb, 7, 7) |
            util_bitpack_uint(values->range, 8, 9) |
            util_bitpack_uint(values->has_signed_components, 10, 10) |
            util_bitpack_uint(values->unknown_0, 11, 15) |
            util_bitpack_uint(values->stride, 16, 30) |
            util_bitpack_uint(values->unknown_1, 31, 38);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_1", values->unknown_1, 0x100ull);
   lima_genxml_validate_bounds("Texture Descriptor::unnorm_coords", values->unnorm_coords, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_2", values->unknown_2, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::cube_map", values->cube_map, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::sampler_dim", values->sampler_dim, 0x4ull);
   cl[ 1] = util_bitpack_uint(values->unknown_1, 31, 38) >> 32 |
            util_bitpack_uint(values->unnorm_coords, 7, 7) |
            util_bitpack_uint(values->unknown_2, 8, 8) |
            util_bitpack_uint(values->cube_map, 9, 9) |
            util_bitpack_uint(values->sampler_dim, 10, 11) |
            util_bitpack_ufixed_clamp(values->min_lod, 12, 19, 4) |
            util_bitpack_ufixed_clamp(values->max_lod, 20, 27, 4) |
            util_bitpack_sfixed_clamp(values->lod_bias, 28, 36, 4);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_3", values->unknown_3, 0x8ull);
   lima_genxml_validate_bounds("Texture Descriptor::has_stride", values->has_stride, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::mipfilter", values->mipfilter, 0x4ull);
   lima_genxml_validate_bounds("Texture Descriptor::min_img_filter_nearest", values->min_img_filter_nearest, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::mag_img_filter_nearest", values->mag_img_filter_nearest, 0x2ull);
   lima_genxml_validate_bounds("Texture Descriptor::wrap_s", values->wrap_s, 0x8ull);
   lima_genxml_validate_bounds("Texture Descriptor::wrap_t", values->wrap_t, 0x8ull);
   lima_genxml_validate_bounds("Texture Descriptor::wrap_r", values->wrap_r, 0x8ull);
   lima_genxml_validate_bounds("Texture Descriptor::width", values->width, 0x2000ull);
   cl[ 2] = util_bitpack_sfixed_clamp(values->lod_bias, 28, 36, 4) >> 32 |
            util_bitpack_uint(values->unknown_3, 5, 7) |
            util_bitpack_uint(values->has_stride, 8, 8) |
            util_bitpack_uint(values->mipfilter, 9, 10) |
            util_bitpack_uint(values->min_img_filter_nearest, 11, 11) |
            util_bitpack_uint(values->mag_img_filter_nearest, 12, 12) |
            util_bitpack_uint(values->wrap_s, 13, 15) |
            util_bitpack_uint(values->wrap_t, 16, 18) |
            util_bitpack_uint(values->wrap_r, 19, 21) |
            util_bitpack_uint(values->width, 22, 34);
   lima_genxml_validate_bounds("Texture Descriptor::width", values->width, 0x2000ull);
   lima_genxml_validate_bounds("Texture Descriptor::height", values->height, 0x2000ull);
   lima_genxml_validate_bounds("Texture Descriptor::depth", values->depth, 0x2000ull);
   cl[ 3] = util_bitpack_uint(values->width, 22, 34) >> 32 |
            util_bitpack_uint(values->height, 3, 15) |
            util_bitpack_uint(values->depth, 16, 28) |
            __gen_pack_unorm16(values->border_red, 29, 44);
   cl[ 4] = __gen_pack_unorm16(values->border_red, 29, 44) >> 32 |
            __gen_pack_unorm16(values->border_green, 13, 28) |
            __gen_pack_unorm16(values->border_blue, 29, 44);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_4", values->unknown_4, 0x10000ull);
   cl[ 5] = __gen_pack_unorm16(values->border_blue, 29, 44) >> 32 |
            __gen_pack_unorm16(values->border_alpha, 13, 28) |
            util_bitpack_uint(values->unknown_4, 29, 44);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_4", values->unknown_4, 0x10000ull);
   lima_genxml_validate_bounds("Texture Descriptor::layout", values->layout, 0x4ull);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_5", values->unknown_5, 0x8000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_0", values->mip_0, 0x100000000ull);
   cl[ 6] = util_bitpack_uint(values->unknown_4, 29, 44) >> 32 |
            util_bitpack_uint(values->layout, 13, 14) |
            util_bitpack_uint(values->unknown_5, 15, 29) |
            util_bitpack_uint(values->mip_0, 24, 55);
   lima_genxml_validate_bounds("Texture Descriptor::mip_0", values->mip_0, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_1", values->mip_1, 0x100000000ull);
   cl[ 7] = util_bitpack_uint(values->mip_0, 24, 55) >> 32 |
            util_bitpack_uint(values->mip_1, 18, 49);
   lima_genxml_validate_bounds("Texture Descriptor::mip_1", values->mip_1, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_2", values->mip_2, 0x100000000ull);
   cl[ 8] = util_bitpack_uint(values->mip_1, 18, 49) >> 32 |
            util_bitpack_uint(values->mip_2, 12, 43);
   lima_genxml_validate_bounds("Texture Descriptor::mip_2", values->mip_2, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_3", values->mip_3, 0x100000000ull);
   cl[ 9] = util_bitpack_uint(values->mip_2, 12, 43) >> 32 |
            util_bitpack_uint(values->mip_3, 6, 37);
   lima_genxml_validate_bounds("Texture Descriptor::mip_3", values->mip_3, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_4", values->mip_4, 0x100000000ull);
   cl[10] = util_bitpack_uint(values->mip_3, 6, 37) >> 32 |
            util_bitpack_uint(values->mip_4, 0, 31);
   lima_genxml_validate_bounds("Texture Descriptor::mip_5", values->mip_5 >> 6, 0x4000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_6", values->mip_6, 0x100000000ull);
   cl[11] = util_bitpack_uint(values->mip_5 >> 6, 0, 25) |
            util_bitpack_uint(values->mip_6, 20, 51);
   lima_genxml_validate_bounds("Texture Descriptor::mip_6", values->mip_6, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_7", values->mip_7, 0x100000000ull);
   cl[12] = util_bitpack_uint(values->mip_6, 20, 51) >> 32 |
            util_bitpack_uint(values->mip_7, 14, 45);
   lima_genxml_validate_bounds("Texture Descriptor::mip_7", values->mip_7, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_8", values->mip_8, 0x100000000ull);
   cl[13] = util_bitpack_uint(values->mip_7, 14, 45) >> 32 |
            util_bitpack_uint(values->mip_8, 8, 39);
   lima_genxml_validate_bounds("Texture Descriptor::mip_8", values->mip_8, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_9", values->mip_9, 0x100000000ull);
   cl[14] = util_bitpack_uint(values->mip_8, 8, 39) >> 32 |
            util_bitpack_uint(values->mip_9, 2, 33);
   lima_genxml_validate_bounds("Texture Descriptor::mip_9", values->mip_9, 0x100000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::mip_10", values->mip_10 >> 6, 0x4000000ull);
   lima_genxml_validate_bounds("Texture Descriptor::unknown_6", values->unknown_6, 0x10ull);
   cl[15] = util_bitpack_uint(values->mip_9, 2, 33) >> 32 |
            util_bitpack_uint(values->mip_10 >> 6, 2, 27) |
            util_bitpack_uint(values->unknown_6, 28, 31);
}

#define LIMA_TEXTURE_DESCRIPTOR_LENGTH 64
struct lima_texture_descriptor_packed { uint32_t opaque[16];};
static inline bool
LIMA_TEXTURE_DESCRIPTOR_unpack(FILE_TYPE *fp, CONSTANT_ uint8_t * restrict cl,
                               struct LIMA_TEXTURE_DESCRIPTOR * restrict values)
{
   values->texel_format = (enum lima_texel_format) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 0, 5);
   values->reverse_component_order = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 6, 6);
   values->swap_rb = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 7, 7);
   values->range = (enum lima_texture_range) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 8, 9);
   values->has_signed_components = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 10, 10);
   values->unknown_0 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 11, 15);
   values->stride = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 16, 30);
   values->unknown_1 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 31, 38);
   values->unnorm_coords = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 39, 39);
   values->unknown_2 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 40, 40);
   values->cube_map = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 41, 41);
   values->sampler_dim = (enum lima_sampler_dimension) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 42, 43);
   values->min_lod = __gen_unpack_ulod((CONSTANT_ uint32_t *) cl, 44, 51);
   values->max_lod = __gen_unpack_ulod((CONSTANT_ uint32_t *) cl, 52, 59);
   values->lod_bias = __gen_unpack_slod((CONSTANT_ uint32_t *) cl, 60, 68);
   values->unknown_3 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 69, 71);
   values->has_stride = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 72, 72);
   values->mipfilter = (enum lima_mipfilter) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 73, 74);
   values->min_img_filter_nearest = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 75, 75);
   values->mag_img_filter_nearest = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 76, 76);
   values->wrap_s = (enum lima_tex_wrap) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 77, 79);
   values->wrap_t = (enum lima_tex_wrap) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 80, 82);
   values->wrap_r = (enum lima_tex_wrap) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 83, 85);
   values->width = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 86, 98);
   values->height = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 99, 111);
   values->depth = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 112, 124);
   values->border_red = __gen_unpack_unorm16((CONSTANT_ uint32_t *) cl, 125, 140);
   values->border_green = __gen_unpack_unorm16((CONSTANT_ uint32_t *) cl, 141, 156);
   values->border_blue = __gen_unpack_unorm16((CONSTANT_ uint32_t *) cl, 157, 172);
   values->border_alpha = __gen_unpack_unorm16((CONSTANT_ uint32_t *) cl, 173, 188);
   values->unknown_4 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 189, 204);
   values->layout = (enum lima_texture_layout) __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 205, 206);
   values->unknown_5 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 207, 221);
   values->mip_0 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 222, 247) << 6;
   values->mip_1 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 248, 273) << 6;
   values->mip_2 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 274, 299) << 6;
   values->mip_3 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 300, 325) << 6;
   values->mip_4 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 326, 351) << 6;
   values->mip_5 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 352, 377) << 6;
   values->mip_6 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 378, 403) << 6;
   values->mip_7 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 404, 429) << 6;
   values->mip_8 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 430, 455) << 6;
   values->mip_9 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 456, 481) << 6;
   values->mip_10 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 482, 507) << 6;
   values->unknown_6 = __gen_unpack_uint((CONSTANT_ uint32_t *) cl, 508, 511);
   return true;
}

static inline void
LIMA_TEXTURE_DESCRIPTOR_print(FILE *fp, const struct LIMA_TEXTURE_DESCRIPTOR * values, unsigned indent)
{
   if (lima_texel_format_as_str(values->texel_format))
     fprintf(fp, "%*sTexel Format: %s\n", indent, "", lima_texel_format_as_str(values->texel_format));
   else
     fprintf(fp, "%*sTexel Format: unknown %X (XXX)\n", indent, "", values->texel_format);
   fprintf(fp, "%*sReverse Component Order: %s\n", indent, "", values->reverse_component_order ? "true" : "false");
   fprintf(fp, "%*sSwap RB: %s\n", indent, "", values->swap_rb ? "true" : "false");
   if (lima_texture_range_as_str(values->range))
     fprintf(fp, "%*sRange: %s\n", indent, "", lima_texture_range_as_str(values->range));
   else
     fprintf(fp, "%*sRange: unknown %X (XXX)\n", indent, "", values->range);
   fprintf(fp, "%*sHas Signed Components: %s\n", indent, "", values->has_signed_components ? "true" : "false");
   fprintf(fp, "%*sUnknown 0: 0x%" PRIx32 "\n", indent, "", values->unknown_0);
   fprintf(fp, "%*sStride: %u\n", indent, "", values->stride);
   fprintf(fp, "%*sUnknown 1: 0x%" PRIx32 "\n", indent, "", values->unknown_1);
   fprintf(fp, "%*sUnnorm Coords: %s\n", indent, "", values->unnorm_coords ? "true" : "false");
   fprintf(fp, "%*sUnknown 2: %s\n", indent, "", values->unknown_2 ? "true" : "false");
   fprintf(fp, "%*sCube Map: %s\n", indent, "", values->cube_map ? "true" : "false");
   if (lima_sampler_dimension_as_str(values->sampler_dim))
     fprintf(fp, "%*sSampler Dim: %s\n", indent, "", lima_sampler_dimension_as_str(values->sampler_dim));
   else
     fprintf(fp, "%*sSampler Dim: unknown %X (XXX)\n", indent, "", values->sampler_dim);
   fprintf(fp, "%*sMin Lod: %f\n", indent, "", values->min_lod);
   fprintf(fp, "%*sMax Lod: %f\n", indent, "", values->max_lod);
   fprintf(fp, "%*sLod Bias: %f\n", indent, "", values->lod_bias);
   fprintf(fp, "%*sUnknown 3: 0x%" PRIx32 "\n", indent, "", values->unknown_3);
   fprintf(fp, "%*sHas Stride: %s\n", indent, "", values->has_stride ? "true" : "false");
   if (lima_mipfilter_as_str(values->mipfilter))
     fprintf(fp, "%*sMipfilter: %s\n", indent, "", lima_mipfilter_as_str(values->mipfilter));
   else
     fprintf(fp, "%*sMipfilter: unknown %X (XXX)\n", indent, "", values->mipfilter);
   fprintf(fp, "%*sMin Img Filter Nearest: %s\n", indent, "", values->min_img_filter_nearest ? "true" : "false");
   fprintf(fp, "%*sMag Img Filter Nearest: %s\n", indent, "", values->mag_img_filter_nearest ? "true" : "false");
   if (lima_tex_wrap_as_str(values->wrap_s))
     fprintf(fp, "%*sWrap S: %s\n", indent, "", lima_tex_wrap_as_str(values->wrap_s));
   else
     fprintf(fp, "%*sWrap S: unknown %X (XXX)\n", indent, "", values->wrap_s);
   if (lima_tex_wrap_as_str(values->wrap_t))
     fprintf(fp, "%*sWrap T: %s\n", indent, "", lima_tex_wrap_as_str(values->wrap_t));
   else
     fprintf(fp, "%*sWrap T: unknown %X (XXX)\n", indent, "", values->wrap_t);
   if (lima_tex_wrap_as_str(values->wrap_r))
     fprintf(fp, "%*sWrap R: %s\n", indent, "", lima_tex_wrap_as_str(values->wrap_r));
   else
     fprintf(fp, "%*sWrap R: unknown %X (XXX)\n", indent, "", values->wrap_r);
   fprintf(fp, "%*sWidth: %u\n", indent, "", values->width);
   fprintf(fp, "%*sHeight: %u\n", indent, "", values->height);
   fprintf(fp, "%*sDepth: %u\n", indent, "", values->depth);
   fprintf(fp, "%*sBorder Red: %f\n", indent, "", values->border_red);
   fprintf(fp, "%*sBorder Green: %f\n", indent, "", values->border_green);
   fprintf(fp, "%*sBorder Blue: %f\n", indent, "", values->border_blue);
   fprintf(fp, "%*sBorder Alpha: %f\n", indent, "", values->border_alpha);
   fprintf(fp, "%*sUnknown 4: 0x%" PRIx32 "\n", indent, "", values->unknown_4);
   if (lima_texture_layout_as_str(values->layout))
     fprintf(fp, "%*sLayout: %s\n", indent, "", lima_texture_layout_as_str(values->layout));
   else
     fprintf(fp, "%*sLayout: unknown %X (XXX)\n", indent, "", values->layout);
   fprintf(fp, "%*sUnknown 5: 0x%" PRIx32 "\n", indent, "", values->unknown_5);
   fprintf(fp, "%*sMip 0: 0x%" PRIx32 "\n", indent, "", values->mip_0);
   fprintf(fp, "%*sMip 1: 0x%" PRIx32 "\n", indent, "", values->mip_1);
   fprintf(fp, "%*sMip 2: 0x%" PRIx32 "\n", indent, "", values->mip_2);
   fprintf(fp, "%*sMip 3: 0x%" PRIx32 "\n", indent, "", values->mip_3);
   fprintf(fp, "%*sMip 4: 0x%" PRIx32 "\n", indent, "", values->mip_4);
   fprintf(fp, "%*sMip 5: 0x%" PRIx32 "\n", indent, "", values->mip_5);
   fprintf(fp, "%*sMip 6: 0x%" PRIx32 "\n", indent, "", values->mip_6);
   fprintf(fp, "%*sMip 7: 0x%" PRIx32 "\n", indent, "", values->mip_7);
   fprintf(fp, "%*sMip 8: 0x%" PRIx32 "\n", indent, "", values->mip_8);
   fprintf(fp, "%*sMip 9: 0x%" PRIx32 "\n", indent, "", values->mip_9);
   fprintf(fp, "%*sMip 10: 0x%" PRIx32 "\n", indent, "", values->mip_10);
   fprintf(fp, "%*sUnknown 6: 0x%" PRIx32 "\n", indent, "", values->unknown_6);
}
